; ---------------------------------------------------------------------
; These two macros perform the grunt work involved in measuring the
; processor clock cycle count for a block of code. These macros must
; be used in pairs, and the block of code must be placed in between
; the ctr_begin and ctr_end macro calls. The ctr_end macro returns
; the lowest cycle count that occurs in any single pass through the
; block of code, corrected for the loop overhead, in EAX (actually
; in EDX:EAX, but the cycle count is unlikely to overflow EAX).
;
; The loopcount parameter sets the sample size. Within my limited
; testing 1000 seems to be a reasonable value.
;
; Note that setting the priority parameter to REALTIME_PRIORITY_CLASS
; involves some risk, as it will cause your process to preempt *all*
; other processes, including critical Windows processes. Setting the
; priority parameter to HIGH_PRIORITY_CLASS instead will significantly
; reduce the risk, and in most cases produce the same cycle count.
;
; These macros require a .586 or higher processor directive.
; ---------------------------------------------------------------------

; ---------------------------------------------------------------------
; The original version of this code contained an error that prevented
; the reference loop from looping, so the reference count was not
; reliably the lowest that otherwise would have occurred in any single
; pass through the block of code. This version corrects the problem.
;
; The original version of this code failed to preserve EBX around the
; instructions that warm up the CPUID instruction. This version
; corrects the problem.
; ---------------------------------------------------------------------

; ---------------------------------------------------------------------
; This version adapted for POASM.
; ---------------------------------------------------------------------

IFNDEF NORMAL_PRIORITY_CLASS
  NORMAL_PRIORITY_CLASS     equ 20h
ENDIF
IFNDEF HIGH_PRIORITY_CLASS
  HIGH_PRIORITY_CLASS       equ 80h
ENDIF
IFNDEF REALTIME_PRIORITY_CLASS
  REALTIME_PRIORITY_CLASS   equ 100h
ENDIF

ctr_begin MACRO loopcount:REQ, priority

    LOCAL refLabel

    IFNDEF __ctr__stuff__defined__
        __ctr__stuff__defined__ equ <1>
        .data
            ALIGN 8           ;; Optimal alignment for QWORD
            __ctr__qword__count__     dq 0
            __ctr__qword__overhead__  dq 0
            __ctr__loop__count__      dd 0
            __ctr__loop__counter__    dd 0
        .code
    ENDIF

    mov DWORD PTR __ctr__qword__count__, -1
    mov DWORD PTR __ctr__qword__count__ + 4, -1
    mov DWORD PTR __ctr__qword__overhead__, -1
    mov DWORD PTR __ctr__qword__overhead__ + 4, -1

    mov __ctr__loop__count__, loopcount
    mov __ctr__loop__counter__, loopcount

    IFNB <priority>
        invoke GetCurrentProcess
        invoke SetPriorityClass, eax, priority
    ENDIF

    push ebx                  ;; Preserve EBX around CPUID
    xor eax, eax              ;; Warm up the CPUID instruction
    cpuid
    xor eax, eax
    cpuid
    xor eax, eax
    cpuid
    pop ebx

    invoke Sleep, 0           ;; Start a new time slice

    ALIGN 16                  ;; Optimal loop alignment for P6

  refLabel:                   ;; Reference loop

    push ebx                  ;; Preserve EBX around CPUID
    xor eax, eax              ;; Use same CPUID input value for each call
    cpuid                     ;; Flush pipe & wait for pending ops to finish
    rdtsc                     ;; Read Time Stamp Counter
    pop ebx

    push edx                  ;; Preserve high-order 32 bits of start count
    push eax                  ;; Preserve low-order 32 bits of start count

    push ebx
    xor eax, eax
    cpuid                     ;; Make sure loop setup instructions finish
    pop ebx

    ;;;;;;;;;;

    push ebx
    xor eax, eax
    cpuid                     ;; Make sure loop instructions finish
    rdtsc                     ;; Read end count
    pop ebx

    pop ecx                   ;; Recover low-order 32 bits of start count
    sub eax, ecx              ;; Low-order 32 bits of loop count in EAX
    pop ecx                   ;; Recover high-order 32 bits of start count
    sbb edx, ecx              ;; High-order 32 bits of loop count in EDX
    ;;
    ;; Capture lowest cycle count that occurs in any loop.
    ;;
    cmp edx, DWORD PTR __ctr__qword__overhead__ + 4
    jne @F
    cmp eax, DWORD PTR __ctr__qword__overhead__
  @@:
    jnb @F
    mov DWORD PTR __ctr__qword__overhead__, eax
    mov DWORD PTR __ctr__qword__overhead__ + 4, edx
  @@:
    sub __ctr__loop__counter__, 1
    jnz refLabel              ;; End of reference loop

    mov __ctr__loop__counter__, loopcount

    invoke Sleep, 0           ;; Start a new time slice

    ALIGN 16                  ;; Optimal loop alignment for P6

    ;------------------------------------------------
    ; A LOCAL in this macro will not be visible from
    ; the ctr_end macro, so need to generate a unique
    ; label that will be visible.
    ;------------------------------------------------

    __work__loop__label__ CATSTR __work__loop__label__,x

  __work__loop__label__:      ;; Working loop

    push ebx                  ;; Preserve EBX around CPUID
    xor eax, eax              ;; Use same CPUID input value for each call
    cpuid                     ;; Flush pipe & wait for pending ops to finish
    rdtsc                     ;; Read Time Stamp Counter
    pop ebx

    push edx                  ;; Preserve high-order 32 bits of start count
    push eax                  ;; Preserve low-order 32 bits of start count

    push ebx
    xor eax, eax
    cpuid                     ;; Make sure loop setup instructions finish
    pop ebx

ENDM

ctr_end MACRO

    push ebx
    xor eax, eax
    cpuid                     ;; Make sure loop instructions finish
    rdtsc                     ;; Read end count
    pop ebx

    pop ecx                   ;; Recover low-order 32 bits of start count
    sub eax, ecx              ;; Low-order 32 bits of loop count in EAX
    pop ecx                   ;; Recover high-order 32 bits of start count
    sbb edx, ecx              ;; High-order 32 bits of loop count in EDX
    ;;
    ;; Capture lowest cycle count that occurs in any loop.
    ;;
    cmp edx, DWORD PTR __ctr__qword__count__ + 4
    jne @F
    cmp eax, DWORD PTR __ctr__qword__count__
  @@:
    jnb @F
    mov DWORD PTR __ctr__qword__count__, eax
    mov DWORD PTR __ctr__qword__count__ + 4, edx
  @@:

    sub __ctr__loop__counter__, 1
    jnz __work__loop__label__ ;; End of working loop

    invoke GetCurrentProcess
    invoke SetPriorityClass, eax, NORMAL_PRIORITY_CLASS

    mov eax, DWORD PTR __ctr__qword__count__
    sub eax, DWORD PTR __ctr__qword__overhead__
    mov edx, DWORD PTR __ctr__qword__count__ + 4
    sbb edx, DWORD PTR __ctr__qword__overhead__ + 4

ENDM
